/*  BV2SMV: Tool for Translating QF_BV problems into SMV.
 *
 *  Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai, FMV, JKU Linz.
 *
 *  All rights reserved.
 *
 *  This file is part of BV2SMV.
 *  See COPYING for more information on using this software.
 */

#include "bv.h"
#include "btorsmt2/btormem.h"
#include "helper.h"
#include "assert.h"

bv2smv_term * bv2smv_var_create(bv2smv * bv2smv, char * name, unsigned int width) {
	bv2smv_term * var;
	bv2smv_hashBucket * bucket;

	bucket = bv2smv_hashTable_insert(bv2smv->var_hashTable, name);
	if ((var = bucket->data.asPtr))
		return var;

	BTOR_NEW(bv2smv->mm, var);
	bucket->data.asPtr = var;

	var->width = width;
	var->state = state_initialized;
	var->termType = var_term;
	var->varName = name;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, var);

	return var;
}

static bv2smv_term * bv2smv_tseitinVar_create(bv2smv * bv2smv, unsigned int width) {
	char * varName;
	BTOR_NEWN(bv2smv->mm, varName, 4 + ceillog10(bv2smv->tseitinVar_counter + 1));
	varName[0] = 't';
	varName[1] = 's';
	varName[2] = '_';
	sprintf(varName + 3, "%d", bv2smv->tseitinVar_counter++);

	return bv2smv_var_create(bv2smv, varName, width);
}

static bv2smv_term * bv2smv_tseitinVar_createFor(BtorSMT2ParseOpt * parseOpt, bv2smv_term * term) {
	bv2smv * bv2smv = parseOpt->context;

	bv2smv_term * var = bv2smv_tseitinVar_create(bv2smv, BV2SMV_STRIP_TERM(term) ->width);

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->asserts, bv2smv_term_bvcomp(parseOpt, var, term));

	return var;
}

bv2smv_term * bv2smv_const_create(bv2smv * bv2smv, char * value) {
	bv2smv_term * _const;

	BTOR_NEW(bv2smv->mm, _const);

	_const->width = strlen(value);
	_const->state = state_initialized;
	_const->termType = const_term;
	_const->constValue = value;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, _const);

	return _const;
}

bv2smv_term * bv2smv_opTerm_create(BtorSMT2ParseOpt * parseOpt, bv2smv_opType opType, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(!y || BV2SMV_STRIP_TERM(x)->width == BV2SMV_STRIP_TERM(y)->width)
		ERRORMSG("operands must be of the same bit-width");

	// Dealing with constants and RHS-ops

	// put the const operand to RHS
	if (BV2SMV_STRIP_TERM(x) ->termType == const_term && y && BV2SMV_STRIP_TERM(y) ->termType != const_term)
		SWAP_POINTERS(x, y);

	if (y && BV2SMV_STRIP_TERM(y) ->termType == const_term) { // if the RHS is a const, introduce Tseitin vars
		if (BV2SMV_STRIP_TERM(x) ->termType != var_term)
			x = bv2smv_tseitinVar_createFor(parseOpt, x);
		if (opType != equal_op)
			y = bv2smv_tseitinVar_createFor(parseOpt, y);

	} else if (BV2SMV_IS_RHS_OP(opType)) { // if the op is an RHS-op, introduce Tseitin vars
		if (BV2SMV_STRIP_TERM(x) ->termType != var_term)
			x = bv2smv_tseitinVar_createFor(parseOpt, x);
		if (y && BV2SMV_STRIP_TERM(y) ->termType != var_term)
			y = bv2smv_tseitinVar_createFor(parseOpt, y);
	} else {
		// move the operands which has an RHS-op to the RHS
		if (BV2SMV_HAS_RHS_OP(BV2SMV_STRIP_TERM(x)) && y && !BV2SMV_HAS_RHS_OP(BV2SMV_STRIP_TERM(y)))
			SWAP_POINTERS(x, y);

		// if the RHS has an RHS-op
		if (y && BV2SMV_HAS_RHS_OP(BV2SMV_STRIP_TERM(y))) {
			// if LHS has an RHS-op as well, introduce a Tseitin var for LHS
			if (BV2SMV_HAS_RHS_OP(BV2SMV_STRIP_TERM(x)))
				x = bv2smv_tseitinVar_createFor(parseOpt, x);
			if (opType != equal_op || BV2SMV_STRIP_TERM(x) ->termType != var_term)
				y = bv2smv_tseitinVar_createFor(parseOpt, y);
		}
	}

	// Create new term
	bv2smv * bv2smv = parseOpt->context;
	bv2smv_term * term;
	BTOR_NEW(bv2smv->mm, term);

	term->width = (opType == equal_op ? 1 : BV2SMV_STRIP_TERM(x) ->width);
	term->state = state_initialized;
	term->termType = op_term;
	term->opType = opType;
	term->children[0] = x;
	term->children[1] = y;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, term);

	return term;
}

bv2smv_term * bv2smv_ite_create(BtorSMT2ParseOpt * parseOpt, bv2smv_term * cond, bv2smv_term * x, bv2smv_term * y) {
	bv2smv * bv2smv = parseOpt->context;

	if (BV2SMV_STRIP_TERM(cond) ->termType == op_term)
		cond = bv2smv_tseitinVar_createFor(parseOpt, cond);

	bv2smv_term * ts_consequence = bv2smv_tseitinVar_create(bv2smv, BV2SMV_STRIP_TERM(x) ->width);

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->asserts, bv2smv_term_implies(parseOpt, cond, bv2smv_term_bvcomp(parseOpt, ts_consequence, x)));
	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->asserts, bv2smv_term_implies(parseOpt, bv2smv_term_bvnot(parseOpt, cond), bv2smv_term_bvcomp(parseOpt, ts_consequence, y)));

	return ts_consequence;
}

bv2smv_term * bv2smv_ineq_create(BtorSMT2ParseOpt * parseOpt, bv2smv_opType opType, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(BV2SMV_IS_INEQ_OP(opType));
	ASSERT(!BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)));

	if (BV2SMV_STRIP_TERM(x) ->termType != var_term)
		x = bv2smv_tseitinVar_createFor(parseOpt, x);
	if (BV2SMV_STRIP_TERM(y) ->termType != var_term)
		y = bv2smv_tseitinVar_createFor(parseOpt, y);

// Create new term
	bv2smv * bv2smv = parseOpt->context;
	bv2smv_term * term;
	BTOR_NEW(bv2smv->mm, term);

	term->width = 1;
	term->state = state_initialized;
	term->termType = op_term;
	term->opType = opType;
	term->children[0] = x;
	term->children[1] = y;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, term);

	return term;
}

bv2smv_term * bv2smv_shift_create(BtorSMT2ParseOpt * parseOpt, bv2smv_opType opType, bv2smv_term * x) {
	ASSERT(opType == shl1_op || opType == lshr1_op || opType == ashr1_op);
	ASSERT(!BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)));

	if (BV2SMV_STRIP_TERM(x) ->termType != var_term)
		x = bv2smv_tseitinVar_createFor(parseOpt, x);

// Create new term
	bv2smv * bv2smv = parseOpt->context;
	bv2smv_term * term;
	BTOR_NEW(bv2smv->mm, term);

	term->width = BV2SMV_STRIP_TERM(x) ->width;
	term->state = state_initialized;
	term->termType = op_term;
	term->opType = opType;
	term->children[0] = x;
	term->children[1] = NULL;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, term);

	return term;
}

bv2smv_term * bv2smv_indexing_create(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, unsigned int index) {
	ASSERT(!BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)));

	if (BV2SMV_STRIP_TERM(x) ->termType != var_term)
		x = bv2smv_tseitinVar_createFor(parseOpt, x);

// Create new term
	bv2smv * bv2smv = parseOpt->context;
	bv2smv_term * term;
	BTOR_NEW(bv2smv->mm, term);

	term->width = 1;
	term->state = state_initialized;
	term->termType = op_term;
	term->opType = indexing_op;
	term->children[0] = x;
	term->children[1] = (bv2smv_term*) (unsigned long) index;

	BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->allTermsStack, term);

	return term;
}
