/*  BV2SMV: Tool for Translating QF_BV problems into SMV.
 *
 *  Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai, FMV, JKU Linz.
 *
 *  All rights reserved.
 *
 *  This file is part of BV2SMV.
 *  See COPYING for more information on using this software.
 */

#include "bv2smv.h"
#include "bv2smv.config.h"
#include "smt2.h"
#include "btorsmt2/btorconst.h"
#include "assert.h"
#include "helper.h"
#include <stdio.h>
#include <math.h>

#define DEBUG_LEVEL 0

#define SIGN_STRING(x) (BV2SMV_ISNEGATED(x) ? "!" : "")

// a pair of strings; one is x's sign, the other is x's name
#define SIGNED_VARNAME(x) SIGN_STRING(x),BV2SMV_STRIP_TERM(x)->varName

#define PRINT_CONST(file, mm,c) { \
		char *s = btor_const_to_decimal((mm), (c)); \
		fprintf(file, "%s", s); \
		btor_freestr((mm), s); \
	} \

bv2smv * bv2smv_create() {
	BtorMemMgr * mm = btor_new_mem_mgr();

	bv2smv * bv2smv;
	BTOR_NEW(mm, bv2smv);

	bv2smv->mm = mm;
	bv2smv->input_file = bv2smv->output_file = NULL;
	bv2smv->verbosity = 0;

	bv2smv->var_hashTable = bv2smv_strHashTable_new(bv2smv->mm);
	bv2smv->index_hashTable = bv2smv_uintHashTable_new(bv2smv->mm);
	bv2smv->tseitinVar_counter = 1;
	BV2SMV_INIT_STACK(bv2smv->asserts);

	BV2SMV_INIT_STACK(bv2smv->allTermsStack);

	return bv2smv;
}

void bv2smv_delete_term(BtorMemMgr * mm, bv2smv_term * term) {
	term = BV2SMV_STRIP_TERM(term);

	if (!term)
		return;

	switch (term->termType) {
	case var_term:
		btor_freestr(mm, term->varName);
		break;
	case const_term:
		btor_delete_const(mm, term->constValue);
		break;
	default:
		break;
	}

	BTOR_DELETE(mm, term);
}

void bv2smv_delete(bv2smv * bv2smv) {
	if (bv2smv->input_file)
		fclose(bv2smv->input_file);
	if (bv2smv->output_file)
		fclose(bv2smv->output_file);

	BtorMemMgr * mm = bv2smv->mm;

	bv2smv_hashTable_delete(bv2smv->var_hashTable);

	bv2smv_hashTable_delete(bv2smv->index_hashTable);

	BTOR_RELEASE_STACK(mm, bv2smv->asserts);

	bv2smv_term * term;
	termStack * allTermsStack = &bv2smv->allTermsStack;
	while (!BTOR_EMPTY_STACK(*allTermsStack)) {
		term = BTOR_POP_STACK(*allTermsStack);
		bv2smv_delete_term(mm, term);
	}
	BTOR_RELEASE_STACK(mm, *allTermsStack);

	BTOR_DELETE(mm, bv2smv);

	btor_delete_mem_mgr(mm);
}

void bv2smv_add_atom(bv2smv * bv2smv, bv2smv_term * term) {
	fprintf(bv2smv->output_file, "  atom%u : boolean;\n", (term->atomIdx = bv2smv->cntAtoms++));
}

void bv2smv_term_create_atoms(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term))
		term = BV2SMV_STRIP_TERM(term);

	if (term->state == state_declared)
		return;
	term->state = state_declared;

	switch (term->termType) {
	case var_term:
		break;
	case const_term:
		break;
	case op_term:
		if (term->opType == equal_op || BV2SMV_IS_INEQ_OP(term->opType)) {
			bv2smv_add_atom(bv2smv, term);

			// in case of bvadd/bvsub on the RHS, add a cin variable
			bv2smv_term * child1 = BV2SMV_STRIP_TERM(term->children[1]);
			if (child1->termType == op_term && (child1->opType == add_op || child1->opType == sub_op || child1->opType == neg_op))
				fprintf(bv2smv->output_file, "  atom%u_cin : boolean;\n", term->atomIdx);
		} else if (term->opType == indexing_op) {
			bv2smv_add_atom(bv2smv, term);

			bv2smv_hashTable_insert(bv2smv->index_hashTable, term->children[1]);

			return;
		}

		bv2smv_term_create_atoms(bv2smv, term->children[0]);
		bv2smv_term_create_atoms(bv2smv, term->children[1]);

		break;
	}
}

void bv2smv_create_atoms(bv2smv * bv2smv) {
	bv2smv->cntAtoms = 0;

	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--)
		bv2smv_term_create_atoms(bv2smv, *pAssert);
	fprintf(bv2smv->output_file, "\n");
}

void bv2smv_term_create_bool_expressions_inner(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term)) {
		fprintf(bv2smv->output_file, "! ");
		term = BV2SMV_STRIP_TERM(term);
	}

	switch (term->termType) {
	case var_term:
		fprintf(bv2smv->output_file, "%s", term->varName);
		break;
	case const_term:
		PRINT_CONST(bv2smv->output_file, bv2smv->mm, term->constValue);
		break;
	case op_term:
		fprintf(bv2smv->output_file, "( ");
		bv2smv_term_create_bool_expressions_inner(bv2smv, term->children[0]);
		switch (term->opType) {
		case (and_op):
			fprintf(bv2smv->output_file, " & ");
			break;
		case (or_op):
			fprintf(bv2smv->output_file, " | ");
			break;
		case (equiv_op):
			fprintf(bv2smv->output_file, " <-> ");
			break;
		default:
			break;
		}
		if (term->opType != indexing_op)
			bv2smv_term_create_bool_expressions_inner(bv2smv, term->children[1]);

		fprintf(bv2smv->output_file, " )");
		break;
	}
}

void bv2smv_term_create_bool_expressions_outer(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term))
		term = BV2SMV_STRIP_TERM(term);

	if (term->state == state_defined)
		return;
	term->state = state_defined;

	if (term->termType == op_term) {
		if (term->opType == equal_op || BV2SMV_IS_INEQ_OP(term->opType)) {
			if (term->opType != equal_op)
				return;

			bv2smv_term * RHS = BV2SMV_STRIP_TERM(term->children[1]);
			if (RHS->termType == const_term || BV2SMV_HAS_RHS_OP(RHS))
				return;

			fprintf(bv2smv->output_file, "  atom%u_l_bool := ", term->atomIdx);
			bv2smv_term_create_bool_expressions_inner(bv2smv, term->children[0]);
			fprintf(bv2smv->output_file, ";\n");
			fprintf(bv2smv->output_file, "  atom%u_r_bool := ", term->atomIdx);
			bv2smv_term_create_bool_expressions_inner(bv2smv, term->children[1]);
			fprintf(bv2smv->output_file, ";\n");
		} else if (term->opType != indexing_op) {
			bv2smv_term_create_bool_expressions_outer(bv2smv, term->children[0]);
			bv2smv_term_create_bool_expressions_outer(bv2smv, term->children[1]);
		}
	}
}

void bv2smv_create_bool_expressions(bv2smv * bv2smv) {
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--)
		bv2smv_term_create_bool_expressions_outer(bv2smv, *pAssert);
	fprintf(bv2smv->output_file, "\n");
}

void bv2smv_term_create_bool_equalities(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term))
		term = BV2SMV_STRIP_TERM(term);

	if (term->state == state_assigned1)
		return;
	term->state = state_assigned1;

	if (term->termType == op_term) {
		if (term->opType == equal_op || BV2SMV_IS_INEQ_OP(term->opType)) {
			if (term->opType != equal_op)
				return;

			bv2smv_term * RHS = BV2SMV_STRIP_TERM(term->children[1]);
			if (RHS->termType == const_term || BV2SMV_HAS_RHS_OP(RHS))
				return;

			fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
			fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
			fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", BV2SMV_STRIP_TERM(term->children[0]) ->width - 1, term->atomIdx);
			fprintf(bv2smv->output_file, "    TRUE : atom%u & (atom%u_l_bool = atom%u_r_bool);\n", term->atomIdx, term->atomIdx, term->atomIdx);
			fprintf(bv2smv->output_file, "    esac;\n\n");
		} else if (term->opType != indexing_op) {
			bv2smv_term_create_bool_equalities(bv2smv, term->children[0]);
			bv2smv_term_create_bool_equalities(bv2smv, term->children[1]);
		}
	}
}

void bv2smv_create_bool_equalities(bv2smv * bv2smv) {
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--)
		bv2smv_term_create_bool_equalities(bv2smv, *pAssert);
}

void bv2smv_term_create_const_equalities(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term))
		term = BV2SMV_STRIP_TERM(term);

	if (term->termType != op_term)
		return;

	if (term->state == state_assigned2)
		return;
	term->state = state_assigned2;

	bv2smv_term * LHS = term->children[0];
	bv2smv_term * RHS = term->children[1];
	bv2smv_term * LHS_stripped = BV2SMV_STRIP_TERM(LHS);
	bv2smv_term * RHS_stripped = BV2SMV_STRIP_TERM(RHS);
	unsigned int widthMinus1 = LHS_stripped->width - 1;

// INEQUALITY

	if (BV2SMV_IS_INEQ_OP(term->opType)) {
		if ((term->opType == ule_op) || (term->opType == sle_op))
			fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
		else
			fprintf(bv2smv->output_file, "  init(atom%u) := FALSE;\n", term->atomIdx);

		fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
		fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);

		if ((term->opType == slt_op) || (term->opType == sle_op)) {
			fprintf(bv2smv->output_file, "    counter_1st_eq_%u & (!%s%s & %s%s) : FALSE;\n", widthMinus1, SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS));
			fprintf(bv2smv->output_file, "    counter_1st_eq_%u & (%s%s & !%s%s) : TRUE;\n", widthMinus1, SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS));
		}

		fprintf(bv2smv->output_file, "    (!%s%s & %s%s) : TRUE;\n", SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS));
		fprintf(bv2smv->output_file, "    (%s%s & !%s%s) : FALSE;\n", SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS));
		fprintf(bv2smv->output_file, "    TRUE : atom%u;\n", term->atomIdx);
		fprintf(bv2smv->output_file, "    esac;\n\n");
	}

// EQUALITY

	else if (term->opType == equal_op) {
		if (LHS_stripped->termType != op_term || LHS_stripped->opType != indexing_op) {
			if (RHS_stripped->termType == var_term)
				return;

			if (RHS_stripped->termType == const_term) {
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%s%s = const", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS));
				PRINT_CONST(bv2smv->output_file, bv2smv->mm, RHS_stripped->constValue);
				fprintf(bv2smv->output_file, ");\n");
				fprintf(bv2smv->output_file, "    esac;\n\n");

				return;
			}

			switch (RHS_stripped->opType) {
			case shl1_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := %s!%s%s;\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS));
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u | counter_1st_eq_%u : atom%u;\n", widthMinus1, widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%snext(%s) = %s%s);\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS_stripped->children[0]));
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			case ashr1_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_1st_eq_%u : atom%u & %s(%s%s = %s%s);\n", widthMinus1, term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(RHS_stripped->children[0]),
						SIGNED_VARNAME(LHS));
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%snext(%s) = %s%s);\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(RHS_stripped->children[0]), SIGNED_VARNAME(LHS));
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			case lshr1_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_1st_eq_%u : atom%u & %s!%s%s;\n", widthMinus1, term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS));
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%snext(%s) = %s%s);\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(RHS_stripped->children[0]), SIGNED_VARNAME(LHS));
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			case add_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%s%s = (%s%s != %s%s != atom%u_cin));\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS),
						SIGNED_VARNAME(RHS_stripped->children[0]), SIGNED_VARNAME(RHS_stripped->children[1]), term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				fprintf(bv2smv->output_file, "  init(atom%u_cin) := FALSE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u_cin) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u_cin;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & ((%s%s & %s%s) | (%s%s & atom%u_cin) | (%s%s & atom%u_cin));\n", term->atomIdx, SIGNED_VARNAME(RHS_stripped->children[0]),
						SIGNED_VARNAME(RHS_stripped->children[1]), SIGNED_VARNAME(RHS_stripped->children[0]), term->atomIdx, SIGNED_VARNAME(RHS_stripped->children[1]), term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			case sub_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%s%s = (%s%s != %s%s != atom%u_cin));\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(RHS_stripped->children[0]),
						SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS_stripped->children[1]), term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				fprintf(bv2smv->output_file, "  init(atom%u_cin) := FALSE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u_cin) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u_cin;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & ((%s%s & %s%s) | (%s%s & atom%u_cin) | (%s%s & atom%u_cin));\n", term->atomIdx, SIGNED_VARNAME(LHS),
						SIGNED_VARNAME(RHS_stripped->children[1]), SIGNED_VARNAME(LHS), term->atomIdx, SIGNED_VARNAME(RHS_stripped->children[1]), term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			case neg_op:
				fprintf(bv2smv->output_file, "  init(atom%u) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & %s(%s%s = (!%s%s != atom%u_cin));\n", term->atomIdx, SIGN_STRING(RHS), SIGNED_VARNAME(LHS), SIGNED_VARNAME(RHS_stripped->children[0]),
						term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				fprintf(bv2smv->output_file, "  init(atom%u_cin) := TRUE;\n", term->atomIdx);
				fprintf(bv2smv->output_file, "  next(atom%u_cin) := case\n", term->atomIdx);
				fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u_cin;\n", widthMinus1, term->atomIdx);
				fprintf(bv2smv->output_file, "    TRUE : atom%u & (!%s%s & atom%u_cin);\n", term->atomIdx, SIGNED_VARNAME(RHS_stripped->children[0]), term->atomIdx);
				fprintf(bv2smv->output_file, "    esac;\n\n");

				break;
			default:
				break;
			}
		}
	}

// INDEXING

	else if (term->opType == indexing_op) {
		fprintf(bv2smv->output_file, "  init(atom%u) := FALSE;\n", term->atomIdx);
		fprintf(bv2smv->output_file, "  next(atom%u) := case\n", term->atomIdx);
		fprintf(bv2smv->output_file, "    counter_gt_%u : atom%u;\n", bv2smv->maxWidth - 1, term->atomIdx);
		fprintf(bv2smv->output_file, "    counter_eq_%u : %s%s;\n", (unsigned int) (unsigned long) RHS, SIGNED_VARNAME(LHS));
		fprintf(bv2smv->output_file, "    TRUE : atom%u;\n", term->atomIdx);
		fprintf(bv2smv->output_file, "    esac;\n\n");

		return;
	}

	bv2smv_term_create_const_equalities(bv2smv, LHS);
	bv2smv_term_create_const_equalities(bv2smv, RHS);
}

void bv2smv_create_const_equalities(bv2smv * bv2smv) {
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--)
		bv2smv_term_create_const_equalities(bv2smv, *pAssert);
}

void bv2smv_term_create_spec(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term)) {
		fprintf(bv2smv->output_file, "!");
		term = BV2SMV_STRIP_TERM(term);
	}

	switch (term->termType) {
	case var_term:
		fprintf(bv2smv->output_file, "%s", term->varName);
		break;
	case const_term:
		if (BV2SMV_ISBOOL(term))
			fprintf(bv2smv->output_file, btor_is_one_const(term->constValue) ? "TRUE" : "FALSE");
		else
			PRINT_CONST(bv2smv->output_file, bv2smv->mm, term->constValue);
		break;
	case op_term:
		if (term->opType == equal_op || BV2SMV_IS_INEQ_OP(term->opType) || term->opType == indexing_op) {
			fprintf(bv2smv->output_file, "atom%u", term->atomIdx);
			break;
		}
		fprintf(bv2smv->output_file, "(");
		bv2smv_term_create_spec(bv2smv, term->children[0]);
		switch (term->opType) {
		case (and_op):
			fprintf(bv2smv->output_file, " & ");
			break;
		case (or_op):
			fprintf(bv2smv->output_file, " | ");
			break;
		case (equiv_op):
			fprintf(bv2smv->output_file, " <-> ");
			break;
		default:
			break;
		}
		bv2smv_term_create_spec(bv2smv, term->children[1]);
		fprintf(bv2smv->output_file, ")");
		break;
	}
}

void bv2smv_create_spec(bv2smv * bv2smv) {
	fprintf(bv2smv->output_file, "SPEC\n  AG(!counter_gt_%u | ", bv2smv->maxWidth - 1);
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--) {
		fprintf(bv2smv->output_file, "!");
		bv2smv_term_create_spec(bv2smv, *pAssert);
		if (cntAsserts > 1)
			fprintf(bv2smv->output_file, " | ");
	}
	fprintf(bv2smv->output_file, ")\n");
}

void bv2smv_term_create_constants(bv2smv_term * term, bv2smv_hashTable * const_hashTable) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term))
		term = BV2SMV_STRIP_TERM(term);

	switch (term->termType) {
	case var_term:
		break;
	case const_term:
		bv2smv_hashTable_insert(const_hashTable, term->constValue);
//		bv2smv_hashTable_insert(const_hashTable, (void*) term->constValue);
		break;
	case op_term:
		bv2smv_term_create_constants(term->children[0], const_hashTable);
		if (term->opType != indexing_op)
			bv2smv_term_create_constants(term->children[1], const_hashTable);
		break;
	}
}

void bv2smv_create_constants(bv2smv * bv2smv, bv2smv_hashTable * const_hashTable) {
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--)
		bv2smv_term_create_constants(*pAssert, const_hashTable);

	bv2smv_hashBucket * currBucket;

	unsigned int i, j;
	unsigned int indexValue;
	unsigned int currWidth;

	for (currBucket = const_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
		const char * constValue = currBucket->key;
		if (btor_is_zero_const(constValue)) {
			fprintf(bv2smv->output_file, "  const0 := FALSE;\n\n");
			continue;
		}
		if (btor_is_ones_const(constValue)) {
			fprintf(bv2smv->output_file, "  const");
			PRINT_CONST(bv2smv->output_file, bv2smv->mm, constValue);
			fprintf(bv2smv->output_file, " := TRUE;\n\n");
			continue;
		}
		fprintf(bv2smv->output_file, "  const");
		PRINT_CONST(bv2smv->output_file, bv2smv->mm, constValue);
		fprintf(bv2smv->output_file, " := case\n");

		unsigned int cntBits = strlen(constValue);
		const char * c = constValue + cntBits - 1;
//		cntBits -= btor_get_num_leading_zeros_const(bv2smv->mm, constValue);

		for (i = 0; cntBits; i++, cntBits--, c--)
			if (*c == '1') {
				i & 1 ? fprintf(bv2smv->output_file, "    counter0_val") : fprintf(bv2smv->output_file, "    !counter0_val");
				for (indexValue = i >> 1, currWidth = 2, j = 1; currWidth < bv2smv->maxWidth; indexValue >>= 1, currWidth <<= 1, j++)
					indexValue & 1 ? fprintf(bv2smv->output_file, " & counter%u_val", j) : fprintf(bv2smv->output_file, " & !counter%u_val", j);
				fprintf(bv2smv->output_file, " : TRUE;\n");
			}
		fprintf(bv2smv->output_file, "    TRUE : FALSE;\n");
		fprintf(bv2smv->output_file, "    esac;\n");
		fprintf(bv2smv->output_file, "\n");

		/*		unsigned long int constValue = (unsigned long int) currBucket->key;
		 if (!constValue) {
		 fprintf(bv2smv->output_file, "  const0 := FALSE;\n\n");
		 continue;
		 }
		 fprintf(bv2smv->output_file, "  const%lu := case\n", constValue);
		 for (i = 0; constValue > 0; i++, constValue >>= 1)
		 if (constValue & 1) {
		 i & 1 ? fprintf(bv2smv->output_file, "    counter0_val") : fprintf(bv2smv->output_file, "    !counter0_val");
		 for (indexValue = i >> 1, currWidth = 2, j = 1; currWidth < bv2smv->maxWidth; indexValue >>= 1, currWidth <<= 1, j++)
		 indexValue & 1 ? fprintf(bv2smv->output_file, " & counter%u_val", j) : fprintf(bv2smv->output_file, " & !counter%u_val", j);
		 fprintf(bv2smv->output_file, " : TRUE;\n");
		 }
		 fprintf(bv2smv->output_file, "    TRUE : FALSE;\n");
		 fprintf(bv2smv->output_file, "    esac;\n");
		 fprintf(bv2smv->output_file, "\n");*/
	}
}

void bv2smv_translate(bv2smv * bv2smv, FILE * dest) {
	fprintf(bv2smv->output_file, "MODULE main\n\n");
	fprintf(bv2smv->output_file, "VAR\n");

	bv2smv_hashBucket * currBucket;
	unsigned int maxWidth = 0;
	unsigned int currWidth, currValue, counterWidth;
	unsigned int i;

	bv2smv_hashTable * width_hashTable = bv2smv_uintHashTable_new(bv2smv->mm);

	for (currBucket = bv2smv->var_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
		bv2smv_term * varData = (bv2smv_term *) currBucket->data.asPtr;
		currWidth = varData->width;
		if (currWidth > maxWidth)
			maxWidth = currWidth;

		bv2smv_hashTable_insert(width_hashTable, (void*) (unsigned long) currWidth);
	}
	bv2smv->maxWidth = maxWidth;

	counterWidth = ceil(log2(maxWidth));

	for (i = counterWidth; i; i--)
		fprintf(bv2smv->output_file, "  counter%u_val : boolean;\n", i - 1);
	fprintf(bv2smv->output_file, "\n");

	for (currBucket = width_hashTable->first; currBucket != 0; currBucket = currBucket->next)
		fprintf(bv2smv->output_file, "  counter_gt_%u : boolean;\n", (unsigned int) (unsigned long) currBucket->key - 1);
	fprintf(bv2smv->output_file, "\n");

	for (currBucket = bv2smv->var_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
		const char * varName = (const char *) currBucket->key;
		fprintf(bv2smv->output_file, "  %s : boolean;\n", varName);
	}
	fprintf(bv2smv->output_file, "\n");

	bv2smv_create_atoms(bv2smv);

	fprintf(bv2smv->output_file, "DEFINE\n");

	if (counterWidth) {
		fprintf(bv2smv->output_file, "  counter0_out := counter0_val & (!counter_gt_%u);\n", maxWidth - 1);
		for (i = counterWidth - 1; i; i--)
			fprintf(bv2smv->output_file, "  counter%u_out := counter%u_val & counter%u_out;\n", i, i, i - 1);
		fprintf(bv2smv->output_file, "\n");

		for (currBucket = bv2smv->index_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
			currValue = (unsigned int) (unsigned long) currBucket->key;

			fprintf(bv2smv->output_file, "  counter_eq_%u := (", currValue);
			for (i = 0; i < counterWidth; i++) {
				if (i)
					fprintf(bv2smv->output_file, " & ");
				if (!(currValue & 1))
					fprintf(bv2smv->output_file, "!");
				fprintf(bv2smv->output_file, "counter%u_val", i);

				currValue >>= 1;
			}
			fprintf(bv2smv->output_file, ");\n");
		}
		for (currBucket = width_hashTable->first; currBucket != 0; currBucket = currBucket->next)
			if ((currValue = (unsigned int) (unsigned long) currBucket->key - 1)) {
				fprintf(bv2smv->output_file, "  counter_1st_eq_%u := (", currValue);
				for (i = 0; currValue; i++) {
					if (currValue & 1) {
						fprintf(bv2smv->output_file, "counter%u_val", i);
						currValue >>= 1;
						currValue ? fprintf(bv2smv->output_file, " & ") : fprintf(bv2smv->output_file, ");\n");
					} else
						currValue >>= 1;
				}
			}
		fprintf(bv2smv->output_file, "\n");
	}

	bv2smv_hashTable * const_hashTable = bv2smv_strHashTable_new(bv2smv->mm);
//	bv2smv_hashTable * const_hashTable = bv2smv_ulongHashTable_new(bv2smv->mm);
	bv2smv_create_constants(bv2smv, const_hashTable);

	bv2smv_create_bool_expressions(bv2smv);

	fprintf(bv2smv->output_file, "ASSIGN\n");

	for (currBucket = bv2smv->var_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
		const char * varName = (const char *) currBucket->key;
		fprintf(bv2smv->output_file, "  init(%s) := {FALSE,TRUE};\n", varName);
		if (BV2SMV_ISBOOL((bv2smv_term *) currBucket->data.asPtr))
			fprintf(bv2smv->output_file, "  next(%s) := %s;\n", varName, varName);
		else
			fprintf(bv2smv->output_file, "  next(%s) := {FALSE,TRUE};\n", varName);
		fprintf(bv2smv->output_file, "\n");
	}

	if (counterWidth) {
		fprintf(bv2smv->output_file, "  init(counter0_val) := FALSE;\n");
		fprintf(bv2smv->output_file, "  next(counter0_val) := counter0_val != (!counter_gt_%u);\n", maxWidth - 1);
		fprintf(bv2smv->output_file, "\n");
		for (i = 1; i < counterWidth; i++) {
			fprintf(bv2smv->output_file, "  init(counter%u_val) := FALSE;\n", i);
			fprintf(bv2smv->output_file, "  next(counter%u_val) := counter%u_val != counter%u_out;\n", i, i, i - 1);
			fprintf(bv2smv->output_file, "\n");
		}
	}

	for (currBucket = width_hashTable->first; currBucket != 0; currBucket = currBucket->next) {
		if (!(currValue = (unsigned int) (unsigned long) currBucket->key - 1)) {
			fprintf(bv2smv->output_file, "  init(counter_gt_0) := FALSE;\n");
			fprintf(bv2smv->output_file, "  next(counter_gt_0) := TRUE;\n\n");
		} else {
			fprintf(bv2smv->output_file, "  init(counter_gt_%u) := FALSE;\n", currValue);
			fprintf(bv2smv->output_file, "  next(counter_gt_%u) := counter_gt_%u | counter_1st_eq_%u;\n\n", currValue, currValue, currValue);
		}
	}
	fprintf(bv2smv->output_file, "\n");

	bv2smv_create_bool_equalities(bv2smv);

	bv2smv_create_const_equalities(bv2smv);

	bv2smv_create_spec(bv2smv);

	bv2smv_hashTable_delete(width_hashTable);
	bv2smv_hashTable_delete(const_hashTable);
}

void bv2smv_term_print(bv2smv * bv2smv, bv2smv_term * term) {
	if (!term)
		return;

	if (BV2SMV_ISNEGATED(term)) {
		printf("! ");
		term = BV2SMV_STRIP_TERM(term);
	}

	switch (term->termType) {
	case var_term:
		printf("%s", term->varName);
		break;
	case const_term:
		PRINT_CONST(stdin, bv2smv->mm, term->constValue);
		break;
	case op_term:
		if (term->opType == neg_op)
			printf("-");
		else
			printf("( ");

		bv2smv_term_print(bv2smv, term->children[0]);

		switch (term->opType) {
		case (and_op):
			printf(" & ");
			break;
		case (or_op):
			printf(" | ");
			break;
		case (equiv_op):
			printf(" <=> ");
			break;
		case (equal_op):
			printf(" = ");
			break;
		case (indexing_op):
			printf("[%d]", (unsigned int) (unsigned long) term->children[1]);
			return;
		case (shl1_op):
			printf(" << 1 ");
			break;
		case (ashr1_op):
			printf(" >>a 1 ");
			break;
		case (lshr1_op):
			printf(" >>l 1 ");
			break;
		case (add_op):
			printf(" + ");
			break;
		case (neg_op):
			return;
		case (sub_op):
			printf(" - ");
			break;
		case (ult_op):
			printf(" <u ");
			break;
		case (ule_op):
			printf(" <=u ");
			break;
		case (slt_op):
			printf(" <s ");
			break;
		case (sle_op):
			printf(" <=s ");
			break;
		default:
			break;
		}

		bv2smv_term_print(bv2smv, term->children[1]);

		printf(" )");
		break;
	}
}

void bv2smv_print(bv2smv * bv2smv) {
	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = bv2smv->asserts.start, cntAsserts = BV2SMV_COUNT_STACK(bv2smv->asserts) ; cntAsserts; pAssert++, cntAsserts--) {
		bv2smv_term_print(bv2smv, *pAssert);
		printf("\n");
	}
}

static void parse_smt2(bv2smv * bv2smv) {
	const BtorParserAPI * parser_api = btor_smt2_parser_api();

	BtorSMT2ParseOpt parse_opt;

	parse_opt.context = bv2smv;

	parse_opt.verbosity = 0;
	parse_opt.need_model = 0;
	parse_opt.window = 0;

	parse_opt.get_exp_len = (btorsmt2_int_fun_type) bv2smv_get_bvWidth;

	parse_opt.var_exp = (btorsmt2_var_fun_type) bv2smv_term_var;
	parse_opt.const_exp = (btorsmt2_const_fun_type) bv2smv_term_const;
	parse_opt.true_exp = (btorsmt2_noarg_fun_type) bv2smv_term_true;
	parse_opt.false_exp = (btorsmt2_noarg_fun_type) bv2smv_term_false;
	parse_opt.zero_exp = (btorsmt2_intarg_fun_type) bv2smv_term_zero;

	parse_opt.bvnot_exp = (btorsmt2_unary_fun_type) bv2smv_term_bvnot;
	parse_opt.bvor_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvor;
	parse_opt.bvnor_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvnor;
	parse_opt.bvand_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvand;
	parse_opt.bvnand_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvnand;
	parse_opt.bvxor_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvxor;
	parse_opt.bvxnor_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvxnor;
	parse_opt.implies_exp = (btorsmt2_binary_fun_type) bv2smv_term_implies;
	parse_opt.ite_exp = (btorsmt2_ternary_fun_type) bv2smv_term_ite;

	parse_opt.eq_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvcomp;
	parse_opt.distinct_exp = (btorsmt2_binary_fun_type) bv2smv_term_distinct;
	parse_opt.bvult_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvult;
	parse_opt.bvule_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvule;
	parse_opt.bvugt_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvugt;
	parse_opt.bvuge_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvuge;
	parse_opt.bvslt_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvslt;
	parse_opt.bvsle_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsle;
	parse_opt.bvsgt_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsgt;
	parse_opt.bvsge_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsge;

	parse_opt.bvneg_exp = (btorsmt2_unary_fun_type) bv2smv_term_bvneg;
	parse_opt.bvadd_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvadd;
	parse_opt.bvsub_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsub;
	parse_opt.bvmul_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvmul;
	parse_opt.bvudiv_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvudiv;
	parse_opt.bvurem_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvurem;
	parse_opt.bvsdiv_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsdiv;
	parse_opt.bvsrem_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsrem;
	parse_opt.bvsmod_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvsmod;

	parse_opt.bvshl_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvshl;
	parse_opt.bvlshr_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvlshr;
	parse_opt.bvashr_exp = (btorsmt2_binary_fun_type) bv2smv_term_bvashr;

	parse_opt.concat_exp = (btorsmt2_binary_fun_type) bv2smv_term_concat;
	parse_opt.extract_exp = (btorsmt2_extract_fun_type) bv2smv_term_extract;
	parse_opt.zero_extend_exp = (btorsmt2_extend_fun_type) bv2smv_term_zero_extend;
	parse_opt.sign_extend_exp = (btorsmt2_extend_fun_type) bv2smv_term_sign_extend;

	parse_opt.copy_exp = (btorsmt2_unary_fun_type) bv2smv_copy_term;
	parse_opt.release_exp = (btorsmt2_void_fun_type) bv2smv_release_term;
	parse_opt.is_array = (btorsmt2_int_fun_type) bv2smv_is_array_term;

	BtorParser * parser = parser_api->init(bv2smv->mm, (BtorParseOpt*) &parse_opt);

	BtorCharStack prefix;
	BTOR_INIT_STACK(prefix);
	BtorParseResult result;
	char * errorMsg = parser_api->parse(parser, &prefix, bv2smv->input_file, "bv2smv", &result);
	ASSERT(!errorMsg)
		ERRORMSG("%s", errorMsg);

	bv2smv_term ** pAssert;
	unsigned int cntAsserts;
	for (pAssert = (bv2smv_term**) result.outputs, cntAsserts = result.noutputs; cntAsserts; pAssert++, cntAsserts--)
		 BV2SMV_PUSH_STACK(bv2smv->mm, bv2smv->asserts, *pAssert);

	parser_api->reset(parser);
}

static int parse_cmdline_args(bv2smv * bv2smv, int argc, char **argv) {
	char ** pArg;
	for (pArg = argv + 1, argc--; argc; pArg++, argc--) {
		if (!strcmp(*pArg, "-h") || !strcmp(*pArg, "--help")) {
			printf("usage: bv2smv [<option>...] [<input>]\n\n");
			printf("where <option> is one of the following:\n\n");
			printf("  -h|--help                        print usage information and exit\n");
			printf("  -c|--copyright                   print copyright and exit\n");
			printf("  -V|--version                     print version and exit\n");
			printf("\n");
			printf("  -v|--verbose                     increase verbosity (0 default)\n");
			printf("  -o|--output <file>               set output file\n");
			/*			printf("  -rwl<n>|--rewrite-level<n>       set rewrite level; <n> must be the sum of some of the following values:\n");
			 printf("                                         1 = rewriting based on true/false constant literals\n");
			 printf("                                         2 = rewriting based on inlining literal definitions\n");*/
			return 0;
		}
		if (!strcmp(*pArg, "-c") || !strcmp(*pArg, "--copyright")) {
			printf("Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai\n");
			printf("Institute for Formal Models and Verification\n");
			printf("Johannes Kepler University, Linz, Austria\n");
			return 0;
		}
		if (!strcmp(*pArg, "-V") || !strcmp(*pArg, "--version")) {
			printf("%s\n", BV2SMV_VERSION);
			return 0;
		}
		if (!strcmp(*pArg, "-o") || !strcmp(*pArg, "--output")) {
			ASSERT(!bv2smv->output_file)
				ERRORMSG("at most 1 output file allowed");
			pArg++;
			argc--;
			ASSERT(argc)
				ERRORMSG("no file name after '--output'")
			ASSERT((bv2smv->output_file = fopen(*pArg, "w")))
				ERRORMSG("cannot create file '%s'", *pArg);
			continue;
		}
		/*		if (strstr(*pArg, "-rwl") == *pArg) {
		 bv2epr->rewrite_level = atoi(*pArg + strlen("-rwl"));
		 continue;
		 }
		 if (strstr(*pArg, "--rewrite-level") == *pArg) {
		 bv2epr->rewrite_level = atoi(*pArg + strlen("--rewrite-level"));
		 continue;
		 }*/
		if (!strcmp(*pArg, "-v") || !strcmp(*pArg, "--verbose")) {
			bv2smv->verbosity++;
			continue;
		}

		ASSERT(!bv2smv->input_file)
			ERRORMSG("at most 1 input file allowed");
		ASSERT((bv2smv->input_file = fopen(*pArg, "r")))
			ERRORMSG("cannot open file '%s'", *pArg);
	}

	if (bv2smv->verbosity) {
		BV2SMV_MSG("App initiated:");
		if (!bv2smv->input_file)
			BV2SMV_MSG("    using standard input");
		if (!bv2smv->output_file)
			BV2SMV_MSG("    using standard output");
//		BV2SMV_MSG("    rewrite level is %d", bv2epr->rewrite_level);
		BV2SMV_MSG("    verbosity = %d", bv2smv->verbosity);
	}

	if (!bv2smv->input_file)
		bv2smv->input_file = stdin;
	if (!bv2smv->output_file)
		bv2smv->output_file = stdout;

	return 1;
}

int main(int argc, char **argv) {
	bv2smv * bv2smv = bv2smv_create();

	if (parse_cmdline_args(bv2smv, argc, argv)) {
		parse_smt2(bv2smv);
		if (bv2smv->verbosity)
			BV2SMV_MSG("SMT2 parsed");

		if (bv2smv->verbosity > 1) {
			BV2SMV_MSG("Input in normal form:");
			bv2smv_print(bv2smv);
		}

		bv2smv_translate(bv2smv, stdout);
		if (bv2smv->verbosity)
			BV2SMV_MSG("SMV generated");
	}

	bv2smv_delete(bv2smv);

	return 0;
}
