/*  BV2SMV: Tool for Translating QF_BV problems into SMV.
 *
 *  Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai, FMV, JKU Linz.
 *
 *  All rights reserved.
 *
 *  This file is part of BV2SMV.
 *  See COPYING for more information on using this software.
 */

#ifndef BV2SMV_H_
#define BV2SMV_H_

#include "hash.h"
#include "stack.h"
#include "btorsmt2/btormem.h"
#include <stdio.h>

#define BV2SMV_MSG(msg, ...) { printf("[bv2smv] "); printf(msg, ##__VA_ARGS__); printf("\n"); fflush(stdout); }

typedef enum bv2smv_termType {
	var_term, const_term, op_term
} bv2smv_termType;

#define BV2SMV_INEQ_OP (1 << 3)
#define BV2SMV_RHS_OP (1 << 4)
typedef enum bv2smv_opType {
	and_op = 0,
	or_op = 1,
	equiv_op = 2,
	equal_op = 3,
	indexing_op = 4,
	ult_op = BV2SMV_INEQ_OP,
	ule_op = 1 + BV2SMV_INEQ_OP,
	slt_op = 2 + BV2SMV_INEQ_OP,
	sle_op = 3 + BV2SMV_INEQ_OP,
	shl1_op = BV2SMV_RHS_OP,
	lshr1_op = 1 + BV2SMV_RHS_OP,
	ashr1_op = 2 + BV2SMV_RHS_OP,
	add_op = 3 + BV2SMV_RHS_OP,
	neg_op = 4 + BV2SMV_RHS_OP,
	sub_op = 5 + BV2SMV_RHS_OP
} bv2smv_opType;

// is an operator inequality
#define BV2SMV_IS_INEQ_OP(o) ((o) & BV2SMV_INEQ_OP)

// is an operator an RHS-op (bvshl, bvlshr, bvashr, bvadd, bvneg, bvsub), which means that it has to be moved to the RHS of an equality
#define BV2SMV_IS_RHS_OP(o) ((o) & BV2SMV_RHS_OP)
#define BV2SMV_HAS_RHS_OP(x) ((x)->termType == op_term && BV2SMV_IS_RHS_OP((x)->opType))

#define BV2SMV_NEGATE(x) ((bv2smv_term*)(1 ^ (unsigned long)(x)))
#define BV2SMV_ISNEGATED(x) (1 & (unsigned long)(x))
#define BV2SMV_STRIP_TERM(x) ((bv2smv_term*)(~1 & (unsigned long)(x)))

typedef enum bv2smv_termState {
	state_initialized, state_declared, state_defined, state_assigned1, state_assigned2, state_specified
} bv2smv_termState;

typedef struct bv2smv_term {
	unsigned int width;
	bv2smv_termState state :3;
	bv2smv_termType termType :2;
	union {
		char * varName;
		char * constValue;
		struct {
			bv2smv_opType opType;
			struct bv2smv_term * children[2];
			unsigned int atomIdx;
		};
	};
} bv2smv_term;

// is a term Boolean
#define BV2SMV_ISBOOL(x) ((x)->width == 1)

BV2SMV_DECLARE_STACK(term, bv2smv_term*);

typedef struct bv2smv {
	BtorMemMgr * mm;
	FILE *input_file, *output_file;
	unsigned int verbosity;

	bv2smv_hashTable * var_hashTable;
	bv2smv_hashTable * index_hashTable;

	unsigned int tseitinVar_counter;
	termStack asserts;
	unsigned int maxWidth;
	unsigned int cntAtoms;

	termStack allTermsStack;
} bv2smv;

bv2smv * bv2smv_create();
void bv2smv_delete(bv2smv * bv2smv);
void bv2smv_delete_term(BtorMemMgr * mm, bv2smv_term * term);
void bv2smv_translate(bv2smv * bv2smv, FILE * dest);

#endif /* BV2SMV_H_ */
