/*  BV2SMV: Tool for Translating QF_BV problems into SMV.
 *
 *  Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai, FMV, JKU Linz.
 *
 *  All rights reserved.
 *
 *  This file is part of BV2SMV.
 *  See COPYING for more information on using this software.
 */

#include "hash.h"
#include "btorsmt2/btormem.h"
#include "assert.h"
#include <stdio.h>
#include <string.h>

/*
 * Decide whether a hashtable should considered "full"
 */
#define IS_FULL(count, size) (count >= 0.8*size)

/*
 * Default hash, comparison, and deletion functions
 */
static unsigned hash_fun(const void * p) {
	return 1183477 * (unsigned) (unsigned long) p;
}

static int cmp_fun(const void * p, const void * q) {
	return ((long) p) - ((long) q);
}

/*static void del_ptr(QPRMemMan * mem, void * p) {
 DELETE(mem, p);
 }*/

/*
 * Allocate a new hashtable with given hash and comparison functions
 */
bv2smv_hashTable * bv2smv_hashTable_new(BtorMemMgr * mm, bv2smv_hash_fun hash, bv2smv_hashCmp_fun cmp/*, DeletePtr del*/) {
	bv2smv_hashTable * res;

	// allocate an empty hashtable
	BTOR_CNEW(mm, res);

	res->mem = mm;
	res->hash = hash ? hash : hash_fun;
	res->cmp = cmp ? cmp : cmp_fun;
//	res->del = del ? del : del_ptr;

	return res;
}

/*
 * Deallocate a hashtable
 */
void bv2smv_hashTable_delete(bv2smv_hashTable * p2iht) {
	bv2smv_hashBucket * p, *next;

	for (p = p2iht->first; p; p = next) {
		next = p->next;
// p2iht->del(p2iht->mem, p->key);
//		DELETE(p2iht->mem, p->key);
		BTOR_DELETE(p2iht->mem, p);
	}

	BTOR_DELETEN(p2iht->mem, p2iht->table, p2iht->size);
	BTOR_DELETE(p2iht->mem, p2iht);
}

/*
 * Enlarge a hashtable (when it is full)
 */
static void enlargeHashTable(bv2smv_hashTable * p2iht) {
	bv2smv_hashBucket * p, *chain, **old_table, **new_table;
	unsigned old_size, new_size, i, h;
	bv2smv_hash_fun hash;

	old_size = p2iht->size;
	old_table = p2iht->table;

	new_size = old_size ? 2 * old_size : 1;
	BTOR_CNEWN(p2iht->mem, new_table, new_size);

	hash = p2iht->hash;

	for (i = 0; i < old_size; i++)
		for (p = old_table[i]; p; p = chain) {
			chain = p->chain;
			h = hash(p->key);
			h &= new_size - 1;
			p->chain = new_table[h];
			new_table[h] = p;
		}

	BTOR_DELETEN(p2iht->mem, old_table, old_size);

	p2iht->size = new_size;
	p2iht->table = new_table;
}

/*
 * Find the slot which belongs to a given key
 * If there is no such slot, return NULL
 */
static bv2smv_hashBucket ** findPosInHashTable(bv2smv_hashTable * p2iht, const void * key) {
	bv2smv_hashBucket ** p, *b;
	unsigned h;

//	if (p2iht->count == p2iht->size)
	if (IS_FULL(p2iht->count, p2iht->size))
		enlargeHashTable(p2iht);

	ASSERT(p2iht->size > 0)
		ERRORMSG("the size of a hashtable must be positive");

	h = p2iht->hash(key);
	h &= p2iht->size - 1;

	for (p = p2iht->table + h; (b = *p) && !p2iht->cmp(b->key, key); p = &b->chain)
		;

	return p;
}

bv2smv_hashBucket * bv2smv_hashTable_find(bv2smv_hashTable * p2iht, const void *key) {
	return *findPosInHashTable(p2iht, key);
}

/*
 * Allocate and insert a new slot for a given key
 */
bv2smv_hashBucket * bv2smv_hashTable_insert(bv2smv_hashTable * p2iht, const void * key) {
	bv2smv_hashBucket ** p, *res;

	p = findPosInHashTable(p2iht, key);
	/*	ASSERT(!*p)
	 ERRORMSG("no position found in hashtable");*/

	if (*p)
		return *p;

	BTOR_CNEW(p2iht->mem, res);
	res->key = key;
	*p = res;
	p2iht->count++;

	res->prev = p2iht->last;

	if (p2iht->first)
		p2iht->last->next = res;
	else
		p2iht->first = res;

	p2iht->last = res;

	return res;
}

/*
 * Remove the slot that belongs to a given key
 * \param stored_key_ptr a pointer to the key object in the removed slot, if stored_key_ptr is not NULL
 * \param stored_data_ptr a pointer to the data object in the removed slot, if stored_data_ptr is not NULL
 */
/*void removeFromHashTable(HashTable * table, void * key, void ** stored_key_ptr, HashData * stored_data_ptr) {
 HashBucket ** p, *bucket;

 p = findPosInHashTable(table, key);
 bucket = *p;

 ASSERT(bucket)
 ERRORMSG("no hashbucket returned");

 *p = bucket->chain;

 if (bucket->prev)
 bucket->prev->next = bucket->next;
 else
 table->first = bucket->next;

 if (bucket->next)
 bucket->next->prev = bucket->prev;
 else
 table->last = bucket->prev;

 ASSERT(table->count > 0)
 ERRORMSG("the count of elements in hashtable must be positive before decreasing");

 table->count--;

 if (stored_key_ptr)
 *stored_key_ptr = bucket->key;

 if (stored_data_ptr)
 *stored_data_ptr = bucket->data;

 DELETE(table->mem, bucket);
 }*/

// ******
// HASH TABLE FOR STRINGS
// ******
unsigned int string_hash(const void *key) {
	const char * s = key;

	unsigned int hash = 0;
	int c;

	while ((c = *s++)) {
		//Dan Bernstein uses: hash(i) = hash(i - 1) * 33 ^ str[i]; the magic of number 33 (why it works better than many other constants, prime or not) has never been adequately explained
		hash = ((hash << 5) + hash) ^ c;
		/* hash = hash * 33 ^ c */
	}

	return hash;
}

int string_hashCmp(const void *key1, const void *key2) {
	return !strcmp(key1, key2);
}

bv2smv_hashTable * bv2smv_strHashTable_new(BtorMemMgr * mm) {
	return bv2smv_hashTable_new(mm, string_hash, string_hashCmp);
}

void * bv2smv_strHashTable_find(bv2smv_hashTable * t, const char * bvName) {
	return bv2smv_hashTable_find(t, bvName)->data.asPtr;
}

// ******
// HASH TABLE FOR UNSIGNED INTS
// ******

unsigned int uint_hash(const void *key) {
	return 1183477 * (unsigned int) (unsigned long) key;
}

int uint_hashCmp(const void *key1, const void *key2) {
	return key1 == key2;
}

bv2smv_hashTable * bv2smv_uintHashTable_new(BtorMemMgr * mm) {
	return bv2smv_hashTable_new(mm, uint_hash, uint_hashCmp);
}

// ******
// HASH TABLE FOR UNSIGNED LONGS
// ******

unsigned int ulong_hash(const void *key) {
	return 1183477 * (unsigned long int) key;
}

int ulong_hashCmp(const void *key1, const void *key2) {
	return key1 == key2;
}

bv2smv_hashTable * bv2smv_ulongHashTable_new(BtorMemMgr * mm) {
	return bv2smv_hashTable_new(mm, ulong_hash, ulong_hashCmp);
}

