/*  BV2SMV: Tool for Translating QF_BV problems into SMV.
 *
 *  Copyright (C) 2013 Andreas Fröhlich and Gergely Kovásznai, FMV, JKU Linz.
 *
 *  All rights reserved.
 *
 *  This file is part of BV2SMV.
 *  See COPYING for more information on using this software.
 */

#include "smt2.h"
#include "bv.h"
#include "btorsmt2/btorconst.h"
#include "helper.h"
#include "assert.h"

unsigned int bv2smv_get_bvWidth(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
	return BV2SMV_STRIP_TERM(x) ->width;
}

//=============================================================================================

bv2smv_term * bv2smv_term_var(BtorSMT2ParseOpt * parseOpt, unsigned int bvWidth, const char * bvName) {
	bv2smv * bv2smv = parseOpt->context;

	return bv2smv_var_create(parseOpt->context, btor_strdup(bv2smv->mm, bvName), bvWidth);
}

bv2smv_term * bv2smv_term_const(BtorSMT2ParseOpt * parseOpt, char * str) {
	bv2smv * bv2smv = parseOpt->context;

	return bv2smv_const_create(bv2smv, btor_strdup(bv2smv->mm, str));
}

bv2smv_term * bv2smv_term_true(BtorSMT2ParseOpt * parseOpt) {
	bv2smv * bv2smv = parseOpt->context;

	return bv2smv_const_create(parseOpt->context, btor_one_const(bv2smv->mm, 1));
}

bv2smv_term * bv2smv_term_false(BtorSMT2ParseOpt * parseOpt) {
	bv2smv * bv2smv = parseOpt->context;

	return bv2smv_const_create(parseOpt->context, btor_zero_const(bv2smv->mm, 1));
}

bv2smv_term * bv2smv_term_zero(BtorSMT2ParseOpt * parseOpt, unsigned int bvWidth) {
	bv2smv * bv2smv = parseOpt->context;

	return bv2smv_const_create(parseOpt->context, btor_zero_const(bv2smv->mm, bvWidth));
}

//=============================================================================================

bv2smv_term * bv2smv_term_bvnot(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
	// Boolean case
	bv2smv_term * x_stripped;
	if (BV2SMV_ISBOOL((x_stripped = BV2SMV_STRIP_TERM(x))) && x_stripped->termType == const_term) {
		ASSERT(!BV2SMV_ISNEGATED(x));
		if (btor_is_zero_const(x->constValue))
			return bv2smv_term_true(parseOpt);
		else
			return bv2smv_term_false(parseOpt);
	}

	return BV2SMV_NEGATE(x) ;
}

bv2smv_term * bv2smv_term_bvand(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	bv2smv_term *x_stripped, *y_stripped;
	if (BV2SMV_ISBOOL((x_stripped = BV2SMV_STRIP_TERM(x)))) {
		// move the const operand to the RHS
		if (x_stripped->termType == const_term && BV2SMV_STRIP_TERM(y) ->termType != const_term)
			SWAP_POINTERS(x, y);

		// if const in the RHS
		if ((y_stripped = BV2SMV_STRIP_TERM(y) ) ->termType == const_term) {
			if (btor_is_one_const(y_stripped->constValue))
				return x;
			else
				return y;
		}
	}

	return bv2smv_opTerm_create(parseOpt, and_op, x, y);
}

bv2smv_term * bv2smv_term_bvnand(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvnot(parseOpt, bv2smv_term_bvand(parseOpt, x, y));
}

bv2smv_term * bv2smv_term_bvor(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	bv2smv_term *x_stripped, *y_stripped;
	if (BV2SMV_ISBOOL((x_stripped = BV2SMV_STRIP_TERM(x)))) {
		// move the const operand to the RHS
		if (x_stripped->termType == const_term && BV2SMV_STRIP_TERM(y) ->termType != const_term)
			SWAP_POINTERS(x, y);

		// if const in the RHS
		if ((y_stripped = BV2SMV_STRIP_TERM(y) ) ->termType == const_term) {
			if (btor_is_one_const(y_stripped->constValue))
				return y;
			else
				return x;
		}
	}

	return bv2smv_opTerm_create(parseOpt, or_op, x, y);
}

bv2smv_term * bv2smv_term_bvnor(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvnot(parseOpt, bv2smv_term_bvor(parseOpt, x, y));
}

bv2smv_term * bv2smv_term_bvxor(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvnot(parseOpt, bv2smv_term_bvxnor(parseOpt, x, y));
}

bv2smv_term * bv2smv_term_bvxnor(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	bv2smv_term *x_stripped, *y_stripped;
	if (BV2SMV_ISBOOL((x_stripped = BV2SMV_STRIP_TERM(x)))) {
		// move the const operand to the RHS
		if (x_stripped->termType == const_term && BV2SMV_STRIP_TERM(y) ->termType != const_term)
			SWAP_POINTERS(x, y);

		// if const in the RHS
		if ((y_stripped = BV2SMV_STRIP_TERM(y) ) ->termType == const_term) {
			if (btor_is_one_const(y_stripped->constValue))
				return x;
			else
				return bv2smv_term_bvnot(parseOpt, x);
		}
	}

	return bv2smv_opTerm_create(parseOpt, equiv_op, x, y);
}

bv2smv_term * bv2smv_term_implies(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvor(parseOpt, bv2smv_term_bvnot(parseOpt, x), y);
}

bv2smv_term * bv2smv_term_ite(BtorSMT2ParseOpt * parseOpt, bv2smv_term * cond, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_ite_create(parseOpt, cond, x, y);
}

//=============================================================================================

bv2smv_term * bv2smv_term_bvcomp(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvxnor(parseOpt, x, y);

	return bv2smv_opTerm_create(parseOpt, equal_op, x, y);
}

bv2smv_term * bv2smv_term_distinct(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvnot(parseOpt, bv2smv_term_bvcomp(parseOpt, x, y));
}

bv2smv_term * bv2smv_term_bvult(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvand(parseOpt, bv2smv_term_bvnot(parseOpt, x), y);

	return bv2smv_ineq_create(parseOpt, ult_op, x, y);
}

bv2smv_term * bv2smv_term_bvugt(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvult(parseOpt, y, x);
}

bv2smv_term * bv2smv_term_bvule(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvor(parseOpt, bv2smv_term_bvnot(parseOpt, x), y);

	return bv2smv_ineq_create(parseOpt, ule_op, x, y);
}

bv2smv_term * bv2smv_term_bvuge(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvule(parseOpt, y, x);
}

bv2smv_term * bv2smv_term_bvslt(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvand(parseOpt, x, bv2smv_term_bvnot(parseOpt, y));

	return bv2smv_ineq_create(parseOpt, slt_op, x, y);
}

bv2smv_term * bv2smv_term_bvsgt(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvslt(parseOpt, y, x);
}

bv2smv_term * bv2smv_term_bvsle(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvor(parseOpt, x, bv2smv_term_bvnot(parseOpt, y));

	return bv2smv_ineq_create(parseOpt, sle_op, x, y);
}

bv2smv_term * bv2smv_term_bvsge(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	return bv2smv_term_bvsle(parseOpt, y, x);
}

//=============================================================================================

bv2smv_term * bv2smv_term_bvadd(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvxor(parseOpt, x, y);

	return bv2smv_opTerm_create(parseOpt, add_op, x, y);
}

bv2smv_term * bv2smv_term_bvneg(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return x;

	return bv2smv_opTerm_create(parseOpt, neg_op, x, NULL );
}

bv2smv_term * bv2smv_term_bvsub(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_bvxor(parseOpt, x, y);

	return bv2smv_opTerm_create(parseOpt, sub_op, x, y);
}

bv2smv_term * bv2smv_term_bvmul(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvmul' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_bvudiv(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvudiv' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_bvurem(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvurem' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_bvsdiv(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvsdiv' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_bvsrem(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvsrem' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_bvsmod(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'bvsmod' is not allowed");

	return NULL ;
}

//=============================================================================================

bv2smv_term * bv2smv_term_extract(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, unsigned int indUpper, unsigned int indLower) {
	ASSERT(indUpper == indLower)
		ERRORMSG("'extract' is only allowed in the case of equal indices");

// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return x;

	return bv2smv_indexing_create(parseOpt, x, indUpper);
}

bv2smv_term * bv2smv_term_concat(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(0)
		ERRORMSG("'concat' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_zero_extend(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, unsigned int cntNewBits) {
	ASSERT(0)
		ERRORMSG("'zero_extend' is not allowed");

	return NULL ;
}

bv2smv_term * bv2smv_term_sign_extend(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, unsigned int cntNewBits) {
	ASSERT(0)
		ERRORMSG("'sign_extend' is not allowed");

	return NULL ;
}

//=============================================================================================

bv2smv_term * bv2smv_term_bvshl(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(BV2SMV_STRIP_TERM(y)->termType == const_term && btor_is_one_const(BV2SMV_STRIP_TERM(y)->constValue))
		ERRORMSG("'bvshl' only by 1 is allowed");

// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_false(parseOpt);

	return bv2smv_shift_create(parseOpt, shl1_op, x);
}

bv2smv_term * bv2smv_term_bvlshr(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(BV2SMV_STRIP_TERM(y)->termType == const_term && btor_is_one_const(BV2SMV_STRIP_TERM(y)->constValue))
		ERRORMSG("'bvlshr' only by 1 is allowed");

// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return bv2smv_term_false(parseOpt);

	return bv2smv_shift_create(parseOpt, lshr1_op, x);
}

bv2smv_term * bv2smv_term_bvashr(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x, bv2smv_term * y) {
	ASSERT(BV2SMV_STRIP_TERM(y)->termType == const_term && btor_is_one_const(BV2SMV_STRIP_TERM(y)->constValue))
		ERRORMSG("'bvashr' only by 1 is allowed");

// Boolean case
	if (BV2SMV_ISBOOL(BV2SMV_STRIP_TERM(x)))
		return x;

	return bv2smv_shift_create(parseOpt, ashr1_op, x);
}

//=============================================================================================

bv2smv_term * bv2smv_copy_term(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
	return x;
}

void bv2smv_release_term(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
}

int bv2smv_is_array_term(BtorSMT2ParseOpt * parseOpt, bv2smv_term * x) {
	return 0;
}
